// mul.js

//////////////////////////////////////////////////////////////////////////////////////

Bigint.prototype.twice = function() {
	var table = this.table;
	var len = table.length;
	for( var i = 0; i < len; i++ ) {
		table[ i ] <<= 1;
	}
	Bigint._normalizeTable( table );
}

Bigint.mul = function( number1, number2 ) {
	var bigNumber1 = Bigint.serio( number1 );
	var bigNumber2 = Bigint.serio( number2 );
	if( bigNumber1.isZero() || bigNumber2.isZero() ) return ( new Bigint(0) );
	var bigResult;
	if( bigNumber1.table.length <= __BIGINT_OVERFLOW_FREE_CELLS || bigNumber2.table.length <= __BIGINT_OVERFLOW_FREE_CELLS ) {
		bigResult = Bigint._mul_lite( bigNumber1, bigNumber2 );
	} else {
		bigResult = Bigint._mul( bigNumber1, bigNumber2 );
	}
	if( bigNumber1.isNonNegative !== bigNumber2.isNonNegative ) {
		bigResult.isNonNegative = false;
	}
	return bigResult;
}

Bigint._mul = function( bigNumber1, bigNumber2 ) {
	var table1 = bigNumber1.table, table2 = bigNumber2.table;
	var nLength1 = table1.length, nLength2 = table2.length;
	var nMaxLength = nLength1 + nLength2;
	var tableResult = new Array( nMaxLength );

	for( var i = 0; i < nMaxLength; i++ ) tableResult[ i ] = 0;

	var k;
	for( var i = 0; i < nLength1; i++ ) {
		for( var j = 0; j < nLength2; j++ ) {
			k = i + j;
			tableResult[ k ] += table1[ i ] * table2[ j ];
			if( tableResult[ k ] > __BIGINT_OVERFLOW_PROTECTION_THRESHOLD_MUL ) { // See the header of core.js
				Bigint._tweakOverflow( tableResult, k );
			}
		}
	}
	return Bigint._fromTable( tableResult );
}

Bigint._mul_lite = function( bigNumber1, bigNumber2 ) { // At least one of the two should be 630-digit or less for __BIGINT_RADIX = 7
	var table1 = bigNumber1.table, table2 = bigNumber2.table;
	var nLength1 = table1.length, nLength2 = table2.length;
	var nMaxLength = nLength1 + nLength2;
	var tableResult = new Array( nMaxLength );

	for( var i = 0; i < nMaxLength; i++ ) tableResult[ i ] = 0;

	for( var i = 0; i < nLength1; i++ ) {
		for( var j = 0; j < nLength2; j++ ) {
			tableResult[ i + j ] += table1[ i ] * table2[ j ];
		}
	}
	return Bigint._fromTable( tableResult );
}

Bigint.square = function( input ) {
	var bigNumber = Bigint.serio( input );
	var bigResult;
	if( bigNumber.table.length < 91 ) bigResult = Bigint._square_lite( bigNumber );
	else bigResult = Bigint._square( bigNumber );
	return bigResult;
}

Bigint._square = function( bigNumber ) {
	var table = bigNumber.table;
	var nLength = table.length;
	var nMaxLength = nLength * 2;
	var tableResult = new Array( nMaxLength );
	for( var i = 0; i < nMaxLength; i++ ) tableResult[ i ] = 0;
	var nProduct, k;
	for( var i = 0; i < nLength; i++ ) {
		for( var j = i; j < nLength; j++ ) {
			nProduct = table[ i ] * table[ j ];
			k = i + j;
			tableResult[ k ] += ( i < j )? nProduct * 2 : nProduct ;
			if( tableResult[ k ] > __BIGINT_OVERFLOW_PROTECTION_THRESHOLD_SQ ) { // See the header of core.js
				Bigint._tweakOverflow( tableResult, k );
			}
		}
	}
	return Bigint._fromTable( tableResult );
}

Bigint._square_lite = function( bigNumber ) { // Input should be 630-digit or less
	var table = bigNumber.table;
	var nLength = table.length;
	var nMaxLength = nLength * 2;
	var tableResult = new Array( nMaxLength );
	for( var i = 0; i < nMaxLength; i++ ) tableResult[ i ] = 0;
	var nProduct;
	for( var i = 0; i < nLength; i++ ) {
		for( var j = i; j < nLength; j++ ) {
			nProduct = table[ i ] * table[ j ];
			tableResult[ i + j ] += ( i < j )? nProduct * 2 : nProduct ;
		}
	}
	return Bigint._fromTable( tableResult );
}


Bigint._tweakOverflow = function( table, k ) {
	var tmp = table[ k ];
	var nLow = tmp % __BIGINT_UNIT;
	tmp = ( tmp - nLow ) / __BIGINT_UNIT;

	table[ k ] = nLow;

	var nHigh = tmp % __BIGINT_UNIT;
	var nSuperHigh = ( tmp - nHigh ) / __BIGINT_UNIT;

	if( nHigh ) table[ k + 1 ] += nHigh;
	if( nSuperHigh ) table[ k + 2 ] += nSuperHigh;
}

Bigint.pow = function( base, power ) {
	var bigBase = Bigint.serio( base );
	var bigPower = Bigint.serio( power );

	if( bigPower.isZero() ) return new Bigint(1);
	else if( bigBase.isZero() ) return new Bigint(0);

	if( !bigPower.isNonNegative ) {
		if( bigBase.isEqualTo(1) ) return new Bigint(1);
		else return new Bigint(0);
	}

	// Now the result will be non-zero

	var bigResult = Bigint._pow( bigBase, bigPower );
	if( !bigBase.isNonNegative && bigPower.isEven() ) bigResult.isNonNegative = false;
	return bigResult;
}

Bigint._pow = function( bigBase, bigPower ) {
	var strBinary = bigPower.toBinaryString();
	var lastIndex = strBinary.length - 1;
	var bigWork = bigBase.copy();
	var bigResult = ( strBinary.charAt( lastIndex )  === "1" )? bigBase.copy() : new Bigint(1);
	for( var i = lastIndex - 1; i >= 0; i-- ) {
		bigWork =  (bigWork.length < 91)? Bigint._square_lite( bigWork ) :  Bigint._square( bigWork ) ;
		if( strBinary.charAt(i) === "1" ) {

			bigResult = (bigWork.length < 91)? Bigint._mul_lite( bigResult, bigWork ) : Bigint._mul( bigResult, bigWork ) ;

		}
	}
	return bigResult;
}

Bigint.powmod = function( base, power, modulus ) {
	var bigBase = Bigint.serio( base );
	var bigPower = Bigint.serio( power );
	var bigModulus = Bigint.serio( modulus );

	if( bigModulus.isZero() || !bigModulus.isNonNegative ) {
		return Bigint._error( "powmod", "illegal modulus: " + bigModulus );
	} else if( bigModulus.isEqualTo( 1 ) ) {
		return Bigint._pow( bigBase, bigPower );
	}

	// Now modulus >= 2

	if( bigBase.isZero() ) {
		if( bigPower.isZero() ) return new Bigint( 1 );
		else return new Bigint( 0 );
	}

	if( !bigPower.isNonNegative ) {
		return new Bigint( 0 );
	}
	// Nonzero Minimal
	bigBase = Bigint.mod( bigBase, bigModulus );
	return Bigint._powmod( bigBase, bigPower, bigModulus );
}

Bigint._powmod = function( bigBase, bigPower, bigModulus ) {
	var strBinary = bigPower.toBinaryString();
	var lastIndex = strBinary.length - 1;
	var bigWork = bigBase.copy(); // !!!
	var bigResult = ( strBinary.charAt( lastIndex ) === "1" )? bigBase.copy() : new Bigint(1);
	for( var i = lastIndex - 1; i >= 0; i-- ) {
		bigWork = Bigint._square( bigWork );
		bigWork = Bigint._mod( bigWork, bigModulus );
		if( strBinary.charAt(i) === "1" ) {
			bigResult = Bigint._mul( bigResult, bigWork );
			bigResult = Bigint._mod( bigResult, bigModulus );
		}
	}
	return bigResult;
}

Bigint.factorial = function( input ) {
	if( input === undefined || input === "" ) {
		return Bigint._error( "factorial", "input void" );
	}
	var smallNumber;
	if( input.isBigint ) smallNumber = input.toDecimalString();
	else smallNumber = Number( input );

	if( isNaN( smallNumber ) || Math.floor( smallNumber ) !== smallNumber || smallNumber < 0 ) {
		return Bigint._error( "factorial", "bad input: " + input );
	} else if( smallNumber === 0 ) {
		return new Bigint( 1 );
	}

	var myTable;
	var PADDING;

	if( smallNumber <= 900 ) {
		myTable = Bigint._factorial900( smallNumber );
		PADDING = Bigint._padding13;
	} else if( smallNumber <= 9007 ) {
		myTable = Bigint._factorial_gt900( smallNumber );
		PADDING = Bigint._padding12;
	} else if( smallNumber <= 90071 ) {
		myTable = Bigint._factorial_gt9007( smallNumber );
		PADDING = Bigint._padding11;
	} else {
		return Bigint._error( "factorial" , "input too big: " + input );
	}

	var nLastIndex = myTable.length - 1;
	var strResult = String( myTable[ nLastIndex ] );
	for( var i = nLastIndex - 1; i > -1; i-- ) strResult += PADDING( myTable[ i ] );
	myTable.length = 0, myTable = null;
	return new Bigint( strResult );
}

Bigint._factorial900 = function( nSmallNumber ) {
	var nCalculationParam = 13;
	var UNIT = Math.pow( 10, nCalculationParam );

	var myTable = [ 1 ];
	var nFirstNonzeroItemIndex = 0;
	var nStartIndex;
	var nTableSize, nOverflow, nWork, i;
	for( var n = 2; n <= nSmallNumber; n++ ) {
		nTableSize = myTable.length;
		nOverflow = 0;

		nStartIndex = nFirstNonzeroItemIndex + 1;

		if( myTable[ nFirstNonzeroItemIndex ] ) {
			nWork = myTable[ nFirstNonzeroItemIndex ] * n;
			myTable[ nFirstNonzeroItemIndex ] = nWork % UNIT;
			nOverflow =  ( nWork - myTable[ nFirstNonzeroItemIndex ] ) / UNIT;
		} else {
			nFirstNonzeroItemIndex++;
		}

		for( i = nStartIndex; i < nTableSize; i++ ) {
			nWork = nOverflow + myTable[ i ] * n;
			myTable[ i ] = nWork % UNIT;
			nOverflow =  ( nWork - myTable[ i ] ) / UNIT;
		}
		if( nOverflow ) myTable[ i ] = nOverflow;
	}
	return myTable;
}

Bigint.sqrt = function( input ) {
	var bigNumber = Bigint.serio( input );
	if( bigNumber.isNonNegative ) return Bigint._sqrt( bigNumber );
	else return Bigint._error( "sqrt" , "negative input: " + input );
}

Bigint._sqrt = function( bigNumber ) {
	var strNumber = bigNumber.toDecimalString();
	var nRequiredAccuracy;
	var strFloatInitialValue;

	if( strNumber.length < 16 ) {
		return new Bigint( Math.floor( Math.sqrt( strNumber ) ) );
	} else if( strNumber.length < 309 ) {
		nRequiredAccuracy = Math.ceil( strNumber.length / 2 );
		strFloatInitialValue = String( Math.sqrt( bigNumber.toNumber() ) );
	} else {
		var oInput = bigNumber.toFloatEx();
		nRequiredAccuracy = Math.ceil( oInput.power / 2 );
		if( oInput.power & 1 ) {
			oInput.number *= 1e+15;
			oInput.power -= 15;
		} else {
			oInput.number *= 1e+16;
			oInput.power -= 16;
		}
		strFloatInitialValue = Math.sqrt( oInput.number ) + "e+" + ( oInput.power / 2 );
	}

	var strX = Bigint._fromFloat( strFloatInitialValue ).toDecimalString();

	var nSteps = Math.ceil( Math.log( nRequiredAccuracy / 14 ) / Math.log( 2 ) );
	var nMaxAccuracy = 14 * Math.pow( 2 , nSteps );
	var fBias = nRequiredAccuracy / nMaxAccuracy;
	var nAccuracy = Math.ceil( 14 * fBias );

	var strTmp;
	for( var i=0; i<100; i++ ) {
		nAccuracy *= 2;

document.__BIGINT_HTML_FORM.debug.value += nAccuracy + "...";
defaultStatus = ">>> " + nAccuracy + "...";

	var strA = strNumber.substr( 0 , nAccuracy * 2 );
	var strB = strX.substr( 0 , nAccuracy );
	var strQ = ( nAccuracy < 295 )? Bigint._strdiv_typeF( strA , strB ) : Bigint._strdiv_typeS( strA , strB ) ;
	var nDigits = strNumber.length - strX.length;
	if( Bigint._strcmp2( strA , strB ) >= 0 ) nDigits++;
	var nZeros = nDigits - strQ.length;
	if( nZeros > 0 ) strTmp = strQ + Bigint._getStringZeros( nZeros );
	else strTmp = strQ;

		strTmp = Bigint._add_typeA( strX , strTmp, nAccuracy );
		strX = Bigint._half_typeA( strTmp, nAccuracy );
		if( nAccuracy >= nRequiredAccuracy ) return new Bigint( strX );
	}

	return Bigint._error("_sqrt_typeS", "too many loops");
}

// 2 small numbers to small number
Bigint._mulmod32 = function( int1, int2 ) {
	var al = int1 & 0xFFFF;
	var ah = int1 >>> 16;
	var bl = int2 & 0xFFFF;
	var bh = int2 >>> 16;
	return ( al*bl + (((( al*bh + ah*bl ) & 0xFFFF) << 16 ) >>> 0 ) ) >>> 0;
}

